from typing import Any, Literal, Union

import numpy as np
from numpy.typing import NDArray
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    InstanceOf,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.type_adapter import TypeAdapter
from pydantic.types import NonNegativeInt

# Use `typing_extensions.TypedDict`
from typing_extensions import NotRequired, TypedDict

from .framelib.frame_defs import _mask
from .hw_defs import HwParams
from .ram_types import *

__all__ = ["NeuronDestInfo", "NeuronAttrs"]

L = Literal

# Constant of neuron destination information
TICK_RELATIVE_BIT_MAX = 8  # Unsigned
ADDR_AXON_BIT_MAX = 11  # Unsigned. Use `HwParams.ADDR_AXON_MAX` as the high limit
ADDR_CORE_X_BIT_MAX = 5  # Unsigned
ADDR_CORE_Y_BIT_MAX = 5  # Unsigned
ADDR_CORE_X_RID_BIT_MAX = 5  # Unsigned
ADDR_CORE_Y_RID_BIT_MAX = 5  # Unsigned
ADDR_CHIP_X_BIT_MAX = 5  # Unsigned
ADDR_CHIP_Y_BIT_MAX = 5  # Unsigned

TICK_RELATIVE_MAX = _mask(TICK_RELATIVE_BIT_MAX)
ADDR_AXON_MAX = HwParams.ADDR_AXON_MAX

# Constant of neuron attributes
PRN_SYN_INTGR_BIT_MAX = 1  # Unsigned
PRN_THRES_BIT_MAX = 29  # Unsigned
PRN_LEAK_V_BIT_MAX = 29  # Unsigned
RESET_V_BIT_MAX = 30  # Signed
THRES_MASK_CTRL_BIT_MAX = 5  # Unsigned
NEG_THRES_BIT_MAX = 29  # Unsigned
POS_THRES_BIT_MAX = 29  # Unsigned
LEAK_V_BIT_MAX = 30  # Signed
BIT_TRUNCATE_BIT_MAX = 5  # Unsigned
VJT_PRE_BIT_MAX = 30  # Signed

RESET_V_MAX = _mask(RESET_V_BIT_MAX - 1)
RESET_V_MIN = -(RESET_V_MAX + 1)
THRES_MASK_CTRL_MAX = _mask(THRES_MASK_CTRL_BIT_MAX)
THRES_MASK_CTRL_MIN = 0
NEG_THRES_MAX = _mask(NEG_THRES_BIT_MAX)
NEG_THRES_MIN = 0
POS_THRES_MAX = _mask(POS_THRES_BIT_MAX)
POS_THRES_MIN = 0
LEAK_V_MAX = _mask(LEAK_V_BIT_MAX - 1)  # Only for scalar
LEAK_V_MIN = -(LEAK_V_MAX + 1)  # Only for scalar
BIT_TRUNCATE_MAX = _mask(BIT_TRUNCATE_BIT_MAX)
BIT_TRUNCATE_MIN = 0
VJT_PRE_MAX = _mask(VJT_PRE_BIT_MAX - 1)
VJT_PRE_MIN = -(VJT_PRE_MAX + 1)
VJT_MAX = VJT_PRE_MAX
VJT_MIN = VJT_PRE_MIN


class NeuronDestInfo(BaseModel):
    model_config = ConfigDict(
        extra="ignore", frozen=True, validate_assignment=True, strict=True
    )

    addr_chip_x: NonNegativeInt = Field(
        le=_mask(ADDR_CHIP_X_BIT_MAX), description="X coordinate of the chip."
    )

    addr_chip_y: NonNegativeInt = Field(
        le=_mask(ADDR_CHIP_Y_BIT_MAX), description="Y coordinate of the chip."
    )

    addr_core_x: NonNegativeInt = Field(
        le=_mask(ADDR_CORE_X_BIT_MAX), description="X coordinate of the core."
    )

    addr_core_y: NonNegativeInt = Field(
        le=_mask(ADDR_CORE_Y_BIT_MAX), description="Y coordinate of the core."
    )

    addr_core_x_ex: NonNegativeInt = Field(
        le=_mask(ADDR_CORE_X_RID_BIT_MAX),
        description="X replication identifier bit of the core.",
    )

    addr_core_y_ex: NonNegativeInt = Field(
        le=_mask(ADDR_CORE_Y_RID_BIT_MAX),
        description="Y replication identifier bit of the core.",
    )

    tick_relative: list[InstanceOf[NonNegativeInt]] = Field(
        description="Information of relative ticks.",
    )

    addr_axon: list[InstanceOf[NonNegativeInt]] = Field(
        description="Destination axon address."
    )

    @field_validator("tick_relative")
    @classmethod
    def _tick_relative_check(cls, v):
        if any(tr > _mask(TICK_RELATIVE_BIT_MAX) or tr < 0 for tr in v):
            # DO NOT change the type of exception `ValueError` in the validators below.
            raise ValueError("parameter 'tick relative' out of range.")

        return v

    @field_validator("addr_axon")
    @classmethod
    def _addr_axon_check(cls, v):
        if any(addr > ADDR_AXON_MAX or addr < 0 for addr in v):
            raise ValueError("parameter 'addr_axon' out of range.")

        return v

    @model_validator(mode="after")
    def _length_match_check(self):
        if len(self.tick_relative) != len(self.addr_axon):
            raise ValueError(
                "parameter 'tick relative' & 'addr_axon' must have the same "
                f"length, but {len(self.tick_relative)} != {len(self.addr_axon)}."
            )

        return self


class NeuronAttrs(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
        use_enum_values=True,
        strict=True,
    )

    reset_mode: ResetMode = Field(
        description="Reset mode of neuron.",
    )

    reset_v: int = Field(
        ge=RESET_V_MIN,
        le=RESET_V_MAX,
        description="Reset voltage of membrane potential, 30-bit signed integer.",
    )

    leak_comparison: LeakComparisonMode = Field(
        serialization_alias="leak_post",
        description="Leak after threshold comparison or before.",
    )

    threshold_mask_bits: NonNegativeInt = Field(
        le=THRES_MASK_CTRL_MAX,
        serialization_alias="threshold_mask_ctrl",
        description="X-bits mask for random threshold.",
    )

    neg_thres_mode: NegativeThresholdMode = Field(
        serialization_alias="threshold_neg_mode",
        description="Modes of negative threshold.",
    )

    neg_threshold: NonNegativeInt = Field(
        le=NEG_THRES_MAX,
        serialization_alias="threshold_neg",
        description="Negative threshold, 29-bit unsigned integer.",
    )

    pos_threshold: NonNegativeInt = Field(
        le=POS_THRES_MAX,
        serialization_alias="threshold_pos",
        description="Positive threshold, 29-bit unsigned integer.",
    )

    leak_direction: LeakDirectionMode = Field(
        serialization_alias="leak_reversal_flag",
        description="Direction of leak, forward or reversal.",
    )

    leak_integration_mode: LeakIntegrationMode = Field(
        serialization_alias="leak_det_stoch",
        description="Modes of leak integration, deterministic or stochastic.",
    )

    leak_v: Union[int, NDArray[np.int32]] = Field(
        # ge=LEAK_V_MIN,
        # le=LEAK_V_MAX,
        description="Leak voltage, 30-bit signed integer or a np.int32 array.",
    )

    synaptic_integration_mode: SynapticIntegrationMode = Field(
        serialization_alias="weight_det_stoch",
        description="Modes of synaptic integration, deterministic or stochastic.",
    )

    bit_truncation: NonNegativeInt = Field(
        le=BIT_TRUNCATE_MAX,
        serialization_alias="bit_truncate",
        description="Position of truncation, 5-bit unsigned integer.",
    )

    vjt_init: L[0] = Field(
        default=0,
        description="Initial membrane potential, 30-bit signed integer. Fixed at 0 at initialization.",
    )

    @field_serializer("leak_v", when_used="json")
    def _leak_v(self, leak_v: Union[int, NDArray[np.int32]]) -> Union[int, list[int]]:
        if isinstance(leak_v, np.ndarray):
            return leak_v.tolist()
        else:
            return leak_v


class _NeuronAttrsDict(TypedDict):
    """Typed dictionary of `NeuronAttrs` for typing check."""

    reset_mode: int
    reset_v: int
    leak_post: int
    threshold_mask_ctrl: NonNegativeInt
    threshold_neg_mode: int
    threshold_neg: NonNegativeInt
    threshold_pos: NonNegativeInt
    leak_reversal_flag: int
    leak_det_stoch: int
    # Risky type. However, NDArray[np.int32] cannot be verified in TypedDict.
    leak_v: Union[int, Any]
    weight_det_stoch: int
    bit_truncate: NonNegativeInt
    vjt_init: NotRequired[L[0]]


class _NeuronDestInfoDict(TypedDict):
    """Typed dictionary of `NeuronDestInfo` for typing check."""

    addr_core_x: NonNegativeInt
    addr_core_y: NonNegativeInt
    addr_core_x_ex: NonNegativeInt
    addr_core_y_ex: NonNegativeInt
    addr_chip_x: NonNegativeInt
    addr_chip_y: NonNegativeInt
    tick_relative: list[InstanceOf[NonNegativeInt]]
    addr_axon: list[InstanceOf[NonNegativeInt]]


NeuronAttrsChecker = TypeAdapter(_NeuronAttrsDict)
NeuronDestInfoChecker = TypeAdapter(_NeuronDestInfoDict)



from enum import Enum, IntEnum, unique
from typing import NamedTuple, Sequence

from .coordinate import Coord
from .coordinate import ReplicationId as RId
from .hw_defs import HwParams

__all__ = [
    "RoutingLevel",
    "RoutingDirection",
    "RoutingStatus",
    "RoutingCost",
    "ROUTING_DIRECTIONS_IDX",
    "RoutingCoord",
    "get_routing_consumption",
    "get_multicast_cores",
    "get_replication_id",
]

N_ROUTING_LEVEL = 6  # L0~L5
MAX_ROUTING_PATH_LENGTH = N_ROUTING_LEVEL - 1


@unique
class RoutingLevel(IntEnum):
    """The level of routing node.

    L0-level nodes are leaves of tree to store data. A L0-cluster is a physical core.
    """

    L0 = 0
    L1 = 1
    L2 = 2
    L3 = 3
    L4 = 4
    L5 = 5


@unique
class RoutingDirection(Enum):
    """Indicate the direction of the four children in the cluster."""

    X0Y0 = (0, 0)
    X0Y1 = (0, 1)
    X1Y0 = (1, 0)
    X1Y1 = (1, 1)
    ANY = (-1, -1)

    def to_index(self) -> int:
        """Convert the direction to index in children list."""
        if self is RoutingDirection.ANY:
            raise TypeError("The direction of routing is not specified.")

        x, y = self.value

        if HwParams.COORD_Y_PRIORITY:
            return (x << 1) + y
        else:
            return (y << 1) + x


@unique
class RoutingStatus(IntEnum):
    """Indicate the status of L0-level cluster. Not used."""

    AVAILABLE = 0
    """Available for item to attach."""

    USED = 1
    """An item is attached to this cluster."""

    OCCUPIED = 2
    """Wasted. It will be an optimization goal."""

    ALL_EMPTY = 3
    """Not used."""


class RoutingCost(NamedTuple):
    n_L0: int
    n_L1: int
    n_L2: int
    n_L3: int
    n_L4: int
    n_L5: int = 1

    def get_routing_level(self) -> RoutingLevel:
        """Return the routing cluster level. If the #N of Lx-level > 1, then we need a  \
            cluster with level Lx+1. And we need the #N of routing sub-level clusters.

        XXX: At present, if #N of L5 > 1, raise exception.
        """
        if self.n_L5 > 1:
            raise ValueError(f"#N of L5-level node out of range, got {self.n_L5}.")

        for i in reversed(range(N_ROUTING_LEVEL)):
            if self[i] > 1:
                return RoutingLevel(i + 1)

        return RoutingLevel.L1


ROUTING_DIRECTIONS_IDX = (
    (
        RoutingDirection.X0Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y0,
        RoutingDirection.X1Y1,
    )
    if HwParams.COORD_Y_PRIORITY
    else (
        RoutingDirection.X0Y0,
        RoutingDirection.X1Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y1,
    )
)


class RoutingCoord(NamedTuple):
    """Use router directions to represent the coordinate of a cluster."""

    L4: RoutingDirection = RoutingDirection.ANY
    L3: RoutingDirection = RoutingDirection.ANY
    L2: RoutingDirection = RoutingDirection.ANY
    L1: RoutingDirection = RoutingDirection.ANY
    L0: RoutingDirection = RoutingDirection.ANY

    def _coord_specify_check(self) -> None:
        if RoutingDirection.ANY in self:
            raise ValueError(
                f"The direction of routing is not specified completely, got {self}."
            )

    def _L0_property(self) -> None:
        if self.level > RoutingLevel.L0:
            raise AttributeError(
                f"This property is only for L0-level cluster, but self is {self.level}."
            )

    @property
    def level(self) -> RoutingLevel:
        for i in range(len(self)):
            if self[i] is RoutingDirection.ANY:
                return RoutingLevel(MAX_ROUTING_PATH_LENGTH - i)

        return RoutingLevel.L0

    def to_coord(self) -> Coord:
        self._L0_property()
        self._coord_specify_check()

        x = (
            (self.L4.value[0] << 4)
            + (self.L3.value[0] << 3)
            + (self.L2.value[0] << 2)
            + (self.L1.value[0] << 1)
            + self.L0.value[0]
        )

        y = (
            (self.L4.value[1] << 4)
            + (self.L3.value[1] << 3)
            + (self.L2.value[1] << 2)
            + (self.L1.value[1] << 1)
            + self.L0.value[1]
        )

        return Coord(x, y)


def get_routing_consumption(n_core: int) -> RoutingCost:
    """Get the consumption of clusters at different levels by given the `n_core`."""
    n_sub_node = HwParams.N_SUB_ROUTING_NODE

    # Find the nearest #N(=2^X) to accommodate `n_core` L0-level clusters.
    n_Lx = [0] * N_ROUTING_LEVEL
    n_Lx[0] = 1 << (n_core - 1).bit_length()

    for i in range(N_ROUTING_LEVEL - 1):
        n_Lx[1 + i] = 1 if n_Lx[i] < n_sub_node else (n_Lx[i] // n_sub_node)

    return RoutingCost(*n_Lx)


def get_replication_id(coords: Sequence[Coord]) -> RId:
    """Get the replication ID by given the coordinates.

    Args:
        - coords: sequence of coordinates.
    """
    if len(coords) < 1:
        raise ValueError("the length of coordinates must be at least 1.")

    base_coord = coords[0]
    rid = RId(0, 0)

    for coord in coords[1:]:
        rid |= base_coord ^ coord

    return rid


def get_multicast_cores(base_coord: Coord, rid: RId) -> set[Coord]:
    cores = set()
    corex = set()
    corey = set()
    temp = set()

    corex.add(base_coord.x)
    corey.add(base_coord.y)

    for lx in range(5):
        if (rid.x >> lx) & 1:
            for x in corex:
                temp.add(x ^ (1 << lx))

            corex = corex.union(temp)
            temp.clear()

        if (rid.y >> lx) & 1:
            for y in corey:
                temp.add(y ^ (1 << lx))

            corey = corey.union(temp)
            temp.clear()

    for x in corex:
        for y in corey:
            cores.add(Coord(x, y))

    return cores



from collections.abc import Sequence
from typing import Any, Union, overload

import numpy as np
from numpy.typing import NDArray

from paicorelib import LCN_EX, Coord, NeuronAttrs, NeuronDestInfo, ParamsReg
from paicorelib import ReplicationId as RId
from paicorelib import WeightPrecision as WP

from .frames import *
from .types import FRAME_DTYPE, DataArrayType, FrameArrayType, IntScalarType

__all__ = ["OfflineFrameGen"]


class OfflineFrameGen:
    """Offline frame generator."""

    @staticmethod
    def gen_config_frame1(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, random_seed: IntScalarType
    ) -> OfflineConfigFrame1:
        return OfflineConfigFrame1(chip_coord, core_coord, rid, int(random_seed))

    @overload
    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: ParamsReg
    ) -> OfflineConfigFrame2: ...

    @overload
    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: dict[str, Any]
    ) -> OfflineConfigFrame2: ...

    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: Union[ParamsReg, dict[str, Any]],
    ) -> OfflineConfigFrame2:
        if isinstance(params_reg, ParamsReg):
            _params_reg = params_reg.model_dump(by_alias=True)
        else:
            _params_reg = params_reg

        return OfflineConfigFrame2(chip_coord, core_coord, rid, _params_reg)

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_neuron: int,
        attrs: NeuronAttrs,
        dest_info: NeuronDestInfo,
        lcn_ex: LCN_EX,
        weight_precision: WP,
    ) -> OfflineConfigFrame3: ...

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_neuron: int,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        lcn_ex: LCN_EX,
        weight_precision: WP,
    ) -> OfflineConfigFrame3: ...

    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_neuron: int,
        attrs: Union[NeuronAttrs, dict[str, Any]],
        dest_info: Union[NeuronDestInfo, dict[str, Any]],
        lcn_ex: LCN_EX,
        weight_precision: WP,
    ) -> OfflineConfigFrame3:
        if isinstance(attrs, NeuronAttrs):
            _attrs = attrs.model_dump(by_alias=True)
        else:
            _attrs = attrs

        if isinstance(dest_info, NeuronDestInfo):
            _dest_info = dest_info.model_dump(by_alias=True)
        else:
            _dest_info = dest_info

        return OfflineConfigFrame3(
            chip_coord,
            core_coord,
            rid,
            sram_start_addr,
            n_neuron,
            _attrs,
            _dest_info,
            repeat=(1 << lcn_ex) * (1 << weight_precision),
        )

    @staticmethod
    def gen_config_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OfflineConfigFrame4:
        return OfflineConfigFrame4(
            chip_coord, core_coord, rid, sram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_magic_init_frame(
        chip_coord: Coord,
        core_coord: Union[Coord, Sequence[Coord]],
        redundant_init: bool = True,
    ) -> Union[FrameArrayType, tuple[FrameArrayType, FrameArrayType]]:
        """Magic initialization frames for PAICORE 2.0. DO NOT MODIFY!

        Args:
            - chip_coord: coordinate of the target chip.
            - core_coord: coordinates of the target cores.
            - redundant_init: whether to use redundant initialization frames, in case of failure.

        If use redundant initialization frames, the magic frames are composed of:
            1. [config1[0] of core #1] + [init frame] + [config1[0] of core #2] + [init frame] + ...\
                + [config1[0] of core #N] + [init frame]

            2. [config1[1] of core #1] + [config1[1] of core #2] + [config1[1] of core #2] + ... +  \
                [config1[2] of core #2] + [config1[1] of core #N] + [config1[2] of core #N]
            3. [work1[0] of core #1] + [work1[0] of core #2] + ... + [work1[0] of core #N]

        Else,
            1. [config1[0] of core #1] + [config1[0] of core #2] + ... + [config1[0] of core #N] +  \
                [init frame]

            2, 3 remain the same.

        Returns: two parts of magic frames.
        """
        if isinstance(core_coord, Coord):
            _core_coord = (core_coord,)
        else:
            _core_coord = core_coord

        magic_frame_cf_1 = []
        magic_frame_cf_2 = []
        magic_frame_wf = []
        init_frame = OfflineWorkFrame4(chip_coord)

        for coord in _core_coord:
            config1 = OfflineConfigFrame1(chip_coord, coord, RId(0, 0), 0)
            work1 = OfflineWorkFrame1(chip_coord, coord, RId(0, 0), 0, 0, 0)

            magic_frame_cf_1.append(config1.value[0])
            if redundant_init:
                magic_frame_cf_1.append(init_frame.value[0])

            magic_frame_cf_2.extend((config1.value[1], config1.value[2]))
            magic_frame_wf.append(work1.value[0])

        if not redundant_init:
            magic_frame_cf_1.append(init_frame.value[0])

        magic_frame_cf_2.extend(magic_frame_wf)

        return np.asarray(magic_frame_cf_1, dtype=FRAME_DTYPE), np.asarray(
            magic_frame_cf_2, dtype=FRAME_DTYPE
        )

    @staticmethod
    def gen_testin_frame1(
        chip_coord: Coord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame1:
        return OfflineTestInFrame1(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame1(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        random_seed: IntScalarType,
    ) -> OfflineTestOutFrame1:
        return OfflineTestOutFrame1(test_chip_coord, core_coord, rid, int(random_seed))

    @staticmethod
    def gen_testin_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame2:
        return OfflineTestInFrame2(chip_coord, core_coord, rid)

    @overload
    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: ParamsReg
    ) -> OfflineTestOutFrame2: ...

    @overload
    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: dict[str, Any],
    ) -> OfflineTestOutFrame2: ...

    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: Union[ParamsReg, dict[str, Any]],
    ) -> OfflineTestOutFrame2:
        if isinstance(params_reg, ParamsReg):
            pr = params_reg.model_dump(by_alias=True)
        else:
            pr = params_reg

        return OfflineTestOutFrame2(test_chip_coord, core_coord, rid, pr)

    @staticmethod
    def gen_testin_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        data_package_num: int,
    ) -> OfflineTestInFrame3:
        return OfflineTestInFrame3(
            chip_coord, core_coord, rid, sram_start_addr, data_package_num
        )

    @staticmethod
    def gen_testout_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_neuron: int,
        attrs: Union[NeuronAttrs, dict[str, Any]],
        dest_info: Union[NeuronDestInfo, dict[str, Any]],
        *,
        lcn_ex: LCN_EX = LCN_EX.LCN_1X,
        weight_precision: WP = WP.WEIGHT_WIDTH_1BIT,
    ) -> OfflineTestOutFrame3:
        if isinstance(attrs, NeuronAttrs):
            _attrs = attrs.model_dump(by_alias=True)
        else:
            _attrs = attrs

        if isinstance(dest_info, NeuronDestInfo):
            _dest_info = dest_info.model_dump(by_alias=True)
        else:
            _dest_info = dest_info

        return OfflineTestOutFrame3(
            chip_coord,
            core_coord,
            rid,
            sram_start_addr,
            n_neuron,
            _attrs,
            _dest_info,
            repeat=(1 << lcn_ex) * (1 << weight_precision),
        )

    @staticmethod
    def gen_testin_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        data_package_num: int,
    ) -> OfflineTestInFrame4:
        return OfflineTestInFrame4(
            chip_coord, core_coord, rid, sram_start_addr, data_package_num
        )

    @staticmethod
    def gen_testout_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OfflineTestOutFrame4:
        return OfflineTestOutFrame4(
            chip_coord, core_coord, rid, sram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_work_frame1(
        one_input_node: dict[str, Any],
        data: DataArrayType,
    ) -> FrameArrayType:
        """Generate the common part of the input spike frames by given the info \
            of one input node.

        Args:
            - one_input_node: a dictionary of one input node.
        """
        common_frame_dest = OfflineWorkFrame1._frame_dest_reorganized(one_input_node)
        _data = np.asarray(data, dtype=np.uint8)

        return OfflineFrameGen.gen_work_frame1_fast(common_frame_dest, _data)

    @staticmethod
    def gen_work_frame1_fast(
        frame_dest_info: FrameArrayType, data: NDArray[np.uint8]
    ) -> FrameArrayType:
        _max, _min = np.max(data, axis=None), np.min(data, axis=None)

        if _min < np.iinfo(np.uint8).min or _max > np.iinfo(np.uint8).max:
            raise ValueError(f"data out of range np.int8 ({_min}, {_max}).")

        if frame_dest_info.size != data.size:
            raise ValueError(
                f"the size of frame dest info & data are not equal, {frame_dest_info.size} != {data.size}."
            )

        return OfflineWorkFrame1._gen_frame_fast(frame_dest_info, data.flatten())

    @staticmethod
    def gen_work_frame2(chip_coord: Coord, n_sync: IntScalarType) -> OfflineWorkFrame2:
        return OfflineWorkFrame2(chip_coord, int(n_sync))

    @staticmethod
    def gen_work_frame3(chip_coord: Coord) -> OfflineWorkFrame3:
        return OfflineWorkFrame3(chip_coord)

    @staticmethod
    def gen_work_frame4(chip_coord: Coord) -> OfflineWorkFrame4:
        return OfflineWorkFrame4(chip_coord)



from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator
from pydantic.type_adapter import TypeAdapter
from pydantic.types import NonNegativeInt

# Use `typing_extensions.TypedDict`.
from typing_extensions import TypedDict

from .coordinate import Coord
from .framelib.frame_defs import _mask
from .hw_defs import HwParams
from .reg_types import *

__all__ = ["CoreParams", "ParamsReg"]

L = Literal

NUM_DENDRITE_BIT_MAX = 13  # Unsigned
TICK_WAIT_START_BIT_MAX = 15  # Unsigned
TICK_WAIT_END_BIT_MAX = 15  # Unsigned

# Use `HwParams.N_DENDRITE_MAX_ANN` as the high limit
NUM_DENDRITE_MAX = HwParams.N_DENDRITE_MAX_ANN
TICK_WAIT_START_MAX = _mask(TICK_WAIT_START_BIT_MAX)
TICK_WAIT_END_MAX = _mask(TICK_WAIT_END_BIT_MAX)

NUM_DENDRITE_OUT_OF_RANGE_TEXT = (
    "param 'num_dendrite' out of range. When input width is 8-bit in {0} mode, "
    + "the number of dendrites should be no more than {1}."
)


def _num_dendrite_out_of_range_repr(mode: Literal["ANN", "SNN"]) -> str:
    if mode == "ANN":
        max_limit = HwParams.N_DENDRITE_MAX_ANN
    else:
        max_limit = HwParams.N_DENDRITE_MAX_SNN

    return NUM_DENDRITE_OUT_OF_RANGE_TEXT.format(mode, max_limit)


class CoreParams(BaseModel):
    """Parameter model of register parameters listed in Section 2.4.1.

    NOTE: The parameters input in the model are declared in `docs/Table-of-Terms.md`.
    """

    model_config = ConfigDict(
        extra="ignore", validate_assignment=True, use_enum_values=True, strict=True
    )

    name: str = Field(
        frozen=True, description="Name of the physical core.", exclude=True
    )

    weight_precision: WeightPrecisionType = Field(
        frozen=True,
        serialization_alias="weight_width",
        description="Weight precision of crossbar.",
    )

    lcn_extension: LCNExtensionType = Field(
        frozen=True,
        serialization_alias="LCN",
        description="Scale of fan-in extension.",
    )

    input_width_format: InputWidthFormatType = Field(
        frozen=True,
        serialization_alias="input_width",
        description="Format of input spike.",
    )

    spike_width_format: SpikeWidthFormatType = Field(
        frozen=True,
        serialization_alias="spike_width",
        description="Format of output spike.",
    )

    num_dendrite: NonNegativeInt = Field(
        frozen=True,
        le=NUM_DENDRITE_MAX,
        serialization_alias="neuron_num",
        description="The number of used dendrites.",
    )

    max_pooling_en: MaxPoolingEnableType = Field(
        serialization_alias="pool_max",
        description="Enable max pooling or not in 8-bit input format.",
    )

    tick_wait_start: NonNegativeInt = Field(
        frozen=True,
        le=TICK_WAIT_START_MAX,
        description="The core begins to work at #N sync_all. 0 for not starting while 1 for staring forever.",
    )

    tick_wait_end: NonNegativeInt = Field(
        frozen=True,
        le=TICK_WAIT_END_MAX,
        description="The core keeps working within #N sync_all. 0 for not stopping.",
    )

    snn_mode_en: SNNModeEnableType = Field(
        frozen=True, serialization_alias="snn_en", description="Enable SNN mode or not."
    )

    target_lcn: LCNExtensionType = Field(
        frozen=True,
        serialization_alias="target_LCN",
        description="LCN extension of the core.",
    )

    test_chip_addr: Coord = Field(
        frozen=True, description="Destination address of output test frames."
    )

    @model_validator(mode="after")
    def _neuron_num_range_limit(self):
        _core_mode = get_core_mode(
            self.input_width_format, self.spike_width_format, self.snn_mode_en
        )
        if _core_mode.is_snn:
            if self.num_dendrite > HwParams.N_DENDRITE_MAX_SNN:
                raise ValueError(_num_dendrite_out_of_range_repr("SNN"))
        else:
            if self.num_dendrite > HwParams.N_DENDRITE_MAX_ANN:
                raise ValueError(_num_dendrite_out_of_range_repr("ANN"))

        return self

    @model_validator(mode="after")
    def _max_pooling_en_check(self):
        # XXX: If this parameter doesn't affect anything, this check can be removed &
        # set the entire model frozen=True.
        if (
            self.input_width_format is InputWidthFormatType.WIDTH_1BIT
            and self.max_pooling_en is MaxPoolingEnableType.ENABLE
        ):
            self.max_pooling_en = MaxPoolingEnableType.DISABLE

        return self

    @field_serializer("test_chip_addr")
    def _test_chip_addr(self, test_chip_addr: Coord) -> int:
        return test_chip_addr.address


ParamsReg = CoreParams


class _ParamsRegDict(TypedDict):
    """Typed dictionary of `ParamsReg` for typing check."""

    weight_width: int
    LCN: int
    input_width: L[0, 1]
    spike_width: L[0, 1]
    neuron_num: NonNegativeInt
    pool_max: L[0, 1]
    tick_wait_start: NonNegativeInt
    tick_wait_end: NonNegativeInt
    snn_en: L[0, 1]
    target_LCN: NonNegativeInt
    test_chip_addr: NonNegativeInt


ParamsRegChecker = TypeAdapter(_ParamsRegDict)



from importlib.metadata import version

from .coordinate import *
from .framelib import *

# In recent versions, `HwConfig` for external packages remain unchanged.
from .hw_defs import HwParams as HwConfig
from .ram_model import *
from .ram_types import LeakComparisonMode as LCM
from .ram_types import LeakDirectionMode as LDM
from .ram_types import LeakIntegrationMode as LIM
from .ram_types import NegativeThresholdMode as NTM
from .ram_types import ResetMode as RM
from .ram_types import SynapticIntegrationMode as SIM
from .ram_types import ThresholdMode as TM
from .reg_model import *
from .reg_types import CoreMode
from .reg_types import InputWidthFormatType as InputWidthFormat
from .reg_types import LCNExtensionType as LCN_EX
from .reg_types import MaxPoolingEnableType as MaxPoolingEnable
from .reg_types import SNNModeEnableType as SNNModeEnable
from .reg_types import SpikeWidthFormatType as SpikeWidthFormat
from .reg_types import WeightPrecisionType as WeightPrecision
from .reg_types import get_core_mode
from .routing_defs import *

try:
    __version__ = version("paicorelib")
except Exception:
    __version__ = None



import sys
from enum import Enum, IntEnum, auto, unique
from functools import wraps
from typing import Any

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

__all__ = [
    "WeightPrecisionType",
    "LCNExtensionType",
    "InputWidthFormatType",
    "SpikeWidthFormatType",
    "MaxPoolingEnableType",
    "SNNModeEnableType",
    "CoreType",
    "CoreMode",
    "get_core_mode",
]

"""
    Type defines of parameters of registers & parameters of neuron RAM.
    See Section 2.4.1 in V2.1 Manual for details.
"""


@unique
class WeightPrecisionType(IntEnum):
    """Weight precision of crossbar. 2-bit.

    - `WEIGHT_WIDTH_XBIT` for X-bit. Default value is `WEIGHT_WIDTH_8BIT`.
    """

    WEIGHT_WIDTH_1BIT = 0
    WEIGHT_WIDTH_2BIT = 1
    WEIGHT_WIDTH_4BIT = 2
    WEIGHT_WIDTH_8BIT = 3  # Default value.


@unique
class LCNExtensionType(IntEnum):
    """Scale of Fan-in extension. 4-bit.

    - X-time LCN extension. Default value is `LCN_1X`.

    NOTE:
    - For `MODE_ANN`, `LCN_1X` = 144x.
    - For `MODE_SNN` or `MODE_BANN`, `LCN_1X` = 1152x.
    """

    LCN_1X = 0  # Default value.
    LCN_2X = 1
    LCN_4X = 2
    LCN_8X = 3
    LCN_16X = 4
    LCN_32X = 5
    LCN_64X = 6


@unique
class InputWidthFormatType(IntEnum):
    """Format of input spike. 1-bit.

    - `WIDTH_1BIT`: 1-bit spike. Default value.
    - `WIDTH_8BIT`: 8-bit activation.
    """

    WIDTH_1BIT = 0  # Default value.
    WIDTH_8BIT = 1


@unique
class SpikeWidthFormatType(IntEnum):
    """Format of output spike. 1-bit.

    - `WIDTH_1BIT`: 1-bit spike. Default value.
    - `WIDTH_8BIT`: 8-bit activation.
    """

    WIDTH_1BIT = 0  # Default value.
    WIDTH_8BIT = 1


@unique
class MaxPoolingEnableType(IntEnum):
    """Enable max pooling or not in 8-bit input format. 1-bit.

    - `MAX_POOLING_DISABLE`: pooling max disable. Default value.
    - `MAX_POOLING_ENABLE`: pooling max enable.
    """

    DISABLE = 0  # Default value.
    ENABLE = 1


@unique
class SNNModeEnableType(IntEnum):
    """Enable SNN mode or not. 1-bit.

    - `SNN_MODE_DISABLE`: SNN mode disable.
    - `SNN_MODE_ENABLE`: SNN mode enable. Default value.
    """

    DISABLE = 0
    ENABLE = 1  # Default value.


@unique
class CoreType(Enum):
    """Types of cores."""

    TYPE_OFFLINE = auto()
    TYPE_ONLINE = auto()


_ModeParamTuple: TypeAlias = tuple[
    InputWidthFormatType, SpikeWidthFormatType, SNNModeEnableType
]


@unique
class CoreMode(Enum):
    """Working mode of cores. Decided by `input_width`, `spike_width` and `SNN_EN` of   \
        core parameters registers.

        Mode                        input_width    spike_width    SNN_EN
        BANN                            0               0           0
        SNN                             0               0           1
        BANN/SNN to ANN                 0               1           0
        BANN/SNN to SNN with values     0               1           1
        ANN to BANN/SNN                 1               0      Don't care(0)
        ANN                             1               1      Don't care(0)
    """

    MODE_BANN = (
        InputWidthFormatType.WIDTH_1BIT,
        SpikeWidthFormatType.WIDTH_1BIT,
        SNNModeEnableType.DISABLE,
    )
    MODE_SNN = (
        InputWidthFormatType.WIDTH_1BIT,
        SpikeWidthFormatType.WIDTH_1BIT,
        SNNModeEnableType.ENABLE,
    )
    MODE_BANN_OR_SNN_TO_ANN = (
        InputWidthFormatType.WIDTH_1BIT,
        SpikeWidthFormatType.WIDTH_8BIT,
        SNNModeEnableType.DISABLE,
    )
    MODE_BANN_OR_SNN_TO_SNN = (
        InputWidthFormatType.WIDTH_1BIT,
        SpikeWidthFormatType.WIDTH_8BIT,
        SNNModeEnableType.ENABLE,
    )
    MODE_ANN_TO_BANN_OR_SNN = (
        InputWidthFormatType.WIDTH_8BIT,
        SpikeWidthFormatType.WIDTH_1BIT,
        SNNModeEnableType.DISABLE,
    )
    MODE_ANN = (
        InputWidthFormatType.WIDTH_8BIT,
        SpikeWidthFormatType.WIDTH_8BIT,
        SNNModeEnableType.DISABLE,
    )

    @property
    def is_snn(self) -> bool:
        return self is CoreMode.MODE_SNN or self is CoreMode.MODE_BANN_OR_SNN_TO_SNN

    @property
    def conf(self) -> _ModeParamTuple:
        return self.value


def get_core_mode(
    input_width: InputWidthFormatType,
    spike_width: SpikeWidthFormatType,
    snn_mode: SNNModeEnableType,
) -> CoreMode:
    try:
        return CoreMode((input_width, spike_width, snn_mode))
    except ValueError:
        raise ValueError(
            f"invalid mode conf: (input_width, spike_width, snn_mode) = "
            f"({input_width}, {spike_width}, {snn_mode}).",
        )


def core_mode_check(func):
    @wraps(func)
    def wrapper(reg_dict: dict[str, Any], *args, **kwargs):
        _ = get_core_mode(
            reg_dict["input_width"], reg_dict["spike_width"], reg_dict["snn_en"]
        )
        return func(reg_dict, *args, **kwargs)

    return wrapper
